package net.knarcraft.bookswithoutborders;

import net.knarcraft.bookswithoutborders.command.CommandBooksWithoutBorders;
import net.knarcraft.bookswithoutborders.command.CommandCopy;
import net.knarcraft.bookswithoutborders.command.CommandDecrypt;
import net.knarcraft.bookswithoutborders.command.CommandDelete;
import net.knarcraft.bookswithoutborders.command.CommandDeletePublic;
import net.knarcraft.bookswithoutborders.command.CommandEncrypt;
import net.knarcraft.bookswithoutborders.command.CommandGive;
import net.knarcraft.bookswithoutborders.command.CommandGivePublic;
import net.knarcraft.bookswithoutborders.command.CommandGroupEncrypt;
import net.knarcraft.bookswithoutborders.command.CommandLoad;
import net.knarcraft.bookswithoutborders.command.CommandLoadPublic;
import net.knarcraft.bookswithoutborders.command.CommandReload;
import net.knarcraft.bookswithoutborders.command.CommandSave;
import net.knarcraft.bookswithoutborders.command.CommandSavePublic;
import net.knarcraft.bookswithoutborders.command.CommandSetAuthor;
import net.knarcraft.bookswithoutborders.command.CommandSetBookPrice;
import net.knarcraft.bookswithoutborders.command.CommandSetLore;
import net.knarcraft.bookswithoutborders.command.CommandSetTitle;
import net.knarcraft.bookswithoutborders.command.CommandUnSign;
import net.knarcraft.bookswithoutborders.command.GiveTabCompleter;
import net.knarcraft.bookswithoutborders.listener.PlayerEventListener;
import net.knarcraft.bookswithoutborders.listener.SignEventListener;
import net.knarcraft.bookswithoutborders.utility.BookToFromTextHelper;
import net.knarcraft.bookswithoutborders.utility.EconomyHelper;
import net.knarcraft.bookswithoutborders.utility.FileHelper;
import org.bukkit.Material;
import org.bukkit.OfflinePlayer;
import org.bukkit.command.CommandExecutor;
import org.bukkit.command.CommandSender;
import org.bukkit.command.ConsoleCommandSender;
import org.bukkit.command.PluginCommand;
import org.bukkit.command.TabCompleter;
import org.bukkit.configuration.Configuration;
import org.bukkit.entity.Player;
import org.bukkit.inventory.ItemFactory;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.PlayerInventory;
import org.bukkit.inventory.meta.BookMeta;
import org.bukkit.plugin.PluginManager;
import org.bukkit.plugin.java.JavaPlugin;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import static net.knarcraft.bookswithoutborders.BooksWithoutBordersSettings.getBookFolder;
import static net.knarcraft.bookswithoutborders.BooksWithoutBordersSettings.getErrorColor;
import static net.knarcraft.bookswithoutborders.BooksWithoutBordersSettings.getSlash;
import static net.knarcraft.bookswithoutborders.BooksWithoutBordersSettings.getSuccessColor;
import static net.knarcraft.bookswithoutborders.utility.InputCleaningHelper.cleanString;

public class BooksWithoutBorders extends JavaPlugin {
    public static int bookDuplicateLimit;
    //The separating string between the book title and the book author
    public static String titleAuthorSeparator;
    public static String loreSeparator;
    public static List<String> firstBooks;
    public static String welcomeMessage;
    public static List<String> existingPlayers;
    public static Material bookPriceType = null;
    public static double bookPriceQuantity;
    public static boolean authorOnlyCopy;
    public static boolean useYml;
    public static boolean adminDecrypt;
    private static ItemFactory itemFactory;
    public static Map<String, List<String>> loadList;
    public static BooksWithoutBorders booksWithoutBorders;
    public static ConsoleCommandSender consoleSender;

    @Override
    public void onEnable() {
        booksWithoutBorders = this;
        consoleSender = this.getServer().getConsoleSender();
        loadList = new HashMap<>();
        firstBooks = new ArrayList<>();
        BooksWithoutBordersSettings.initialize(this);

        PluginManager pluginManager = this.getServer().getPluginManager();

        if (getSlash() != null && initialize()) {
            pluginManager.registerEvents(new PlayerEventListener(), this);
            pluginManager.registerEvents(new SignEventListener(), this);
        } else {
            this.getPluginLoader().disablePlugin(this);
        }

        registerCommands();
    }

    public static BooksWithoutBorders getInstance() {
        return booksWithoutBorders;
    }

    /**
     * Registers all commands used by this plugin
     */
    private void registerCommands() {
        registerCommand("give", new CommandGive(), new GiveTabCompleter());
        registerCommand("decrypt", new CommandDecrypt(), null);
        registerCommand("groupEncrypt", new CommandGroupEncrypt(), null);
        registerCommand("delete", new CommandDelete(), null);
        registerCommand("deletePublic", new CommandDeletePublic(), null);
        registerCommand("copy", new CommandCopy(), null);
        registerCommand("unSign", new CommandUnSign(), null);
        registerCommand("encrypt", new CommandEncrypt(), null);
        registerCommand("setBookPrice", new CommandSetBookPrice(), null);
        registerCommand("setLore", new CommandSetLore(), null);
        registerCommand("savePublic", new CommandSavePublic(), null);
        registerCommand("save", new CommandSave(), null);
        registerCommand("setAuthor", new CommandSetAuthor(), null);
        registerCommand("setTitle", new CommandSetTitle(), null);
        registerCommand("load", new CommandLoad(), null);
        registerCommand("loadPublic", new CommandLoadPublic(), null);
        registerCommand("booksWithoutBorders", new CommandBooksWithoutBorders(), null);
        registerCommand("reload", new CommandReload(), null);
        registerCommand("givePublic", new CommandGivePublic(), null);
    }

    /**
     * Registers a command
     *
     * @param commandName <p>The name of the command to register</p>
     * @param executor    <p>The executor to register for the command</p>
     */
    private void registerCommand(String commandName, CommandExecutor executor, TabCompleter tabCompleter) {
        PluginCommand pluginCommand = this.getCommand(commandName);
        if (pluginCommand != null) {
            pluginCommand.setExecutor(executor);
            if (tabCompleter != null) {
                pluginCommand.setTabCompleter(tabCompleter);
            }
        } else {
            sendErrorMessage(consoleSender, "Failed to register command " + commandName);
        }
    }

    private boolean initialize() {
        //initializes Item Factory
        try {
            itemFactory = this.getServer().getItemFactory();
        } catch (java.lang.NoSuchMethodError nsmE) {
            sendErrorMessage(consoleSender, "Warning! [BooksWithoutBorders] failed to initialize!");
            sendErrorMessage(consoleSender, "Please confirm the correct version of [BooksWithoutBorders] is");
            sendErrorMessage(consoleSender, "being run for this version of bukkit!");
            return false;
        }

        //Pulls config values
        if (!loadConfig()) {
            return false;
        }

        //Save config with loaded values to fix invalid config values
        saveConfigValues();

        if (!testFileSaving()) {
            return false;
        }

        return loadExistingPlayers();
    }

    private boolean testFileSaving() {
        File fileTest = new File(getBookFolder());
        File encryptedFileTest = new File(getBookFolder() + "Encrypted" + getSlash());
        if (!fileTest.exists()) {
            try {
                if (!fileTest.mkdir()) {
                    sendErrorMessage(consoleSender, "Saving failed! Aborting...");
                    return false;
                }
            } catch (Exception e) {
                e.printStackTrace();
                return false;
            }
        }
        if (!encryptedFileTest.exists()) {
            try {
                if (!encryptedFileTest.mkdir()) {
                    sendErrorMessage(consoleSender, "Saving failed! Aborting...");
                    return false;
                }
            } catch (Exception e) {
                e.printStackTrace();
                return false;
            }
        }
        return true;
    }

    public void saveConfigValues() {
        Configuration config = this.getConfig();
        config.set("Options.Save_Books_in_Yaml_Format", useYml);
        config.set("Options.Max_Number_of_Duplicates", bookDuplicateLimit);
        config.set("Options.Title-Author_Separator", titleAuthorSeparator);
        config.set("Options.Lore_line_separator", loreSeparator);
        config.set("Options.Books_for_new_players", firstBooks);
        config.set("Options.Message_for_new_players", welcomeMessage);
        if (bookPriceType != null) {
            if (bookPriceType != Material.AIR) {
                config.set("Options.Price_to_create_book.Item_type", bookPriceType.toString());
            } else {
                config.set("Options.Price_to_create_book.Item_type", "Economy");
            }
        } else {
            config.set("Options.Price_to_create_book.Item_type", "Item type name");
        }
        config.set("Options.Price_to_create_book.Required_quantity", bookPriceQuantity);
        config.set("Options.Admin_Auto_Decrypt", adminDecrypt);
        config.set("Options.Author_Only_Copy", authorOnlyCopy);

        //Handles old book and quill settings
        if (config.contains("Options.Require_book_and_quill_to_create_book")) {
            sendSuccessMessage(consoleSender, "[BooksWithoutBorders] Found old config setting \"Require_book_and_quill_to_create_book\"");
            sendSuccessMessage(consoleSender, "Updating to \"Price_to_create_book\" settings");

            if (config.getBoolean("Options.Require_book_and_quill_to_create_book")) {
                bookPriceType = Material.WRITABLE_BOOK;
                bookPriceQuantity = 1;
                config.set("Options.Price_to_create_book.Item_type", bookPriceType.toString());
                config.set("Options.Price_to_create_book.Required_quantity", bookPriceQuantity);
            }

            config.set("Options.Require_book_and_quill_to_create_book", null);
        }
        this.saveConfig();
    }

    public boolean loadConfig() {
        this.reloadConfig();
        Configuration config = this.getConfig();
        try {
            useYml = config.getBoolean("Options.Save_Books_in_Yaml_Format", true);
            bookDuplicateLimit = config.getInt("Options.Max_Number_of_Duplicates", 5);
            titleAuthorSeparator = config.getString("Options.Title-Author_Separator", ",");
            loreSeparator = config.getString("Options.Lore_line_separator", "~");
            adminDecrypt = config.getBoolean("Options.Admin_Auto_Decrypt", false);
            authorOnlyCopy = config.getBoolean("Options.Author_Only_Copy", false);

            //Set books to give new players
            firstBooks = config.getStringList("Options.Books_for_new_players");
            if (config.contains("Options.Book_for_new_players")) {
                firstBooks.add(config.getString("Options.Book_for_new_players"));
            }
            if (firstBooks.isEmpty()) {
                firstBooks.add(" ");
            }

            welcomeMessage = config.getString("Options.Message_for_new_players", " ");

            //Converts string into material
            String sMaterial = config.getString("Options.Price_to_create_book.Item_type", " ");
            if (sMaterial.equalsIgnoreCase("Economy")) {
                if (EconomyHelper.setupEconomy()) {
                    bookPriceType = Material.AIR;
                } else {
                    sendErrorMessage(consoleSender, "BooksWithoutBorders failed to hook into Vault! Book price not set!");
                    bookPriceType = null;
                }
            } else if (!sMaterial.equalsIgnoreCase(" ")) {
                for (Material m : Material.values()) {
                    if (m.toString().equalsIgnoreCase(sMaterial))
                        bookPriceType = m;
                }
            }
            bookPriceQuantity = config.getDouble("Options.Price_to_create_book.Required_quantity", 0);

            //makes sure titleAuthorSeparator is a valid value
            titleAuthorSeparator = cleanString(titleAuthorSeparator);
            if (titleAuthorSeparator.length() != 1) {
                sendErrorMessage(consoleSender, "Title-Author_Separator is set to an invalid value!");
                sendErrorMessage(consoleSender, "Reverting to default value of \",\"");
                titleAuthorSeparator = ",";
                config.set("Options.Title-Author_Separator", titleAuthorSeparator);
            }
        } catch (Exception e) {
            sendErrorMessage(consoleSender, "Warning! Config.yml failed to load!");
            sendErrorMessage(consoleSender, "Try Looking for settings that are missing values!");
            return false;
        }

        return true;
    }

    public boolean loadExistingPlayers() {
        File fTest = new File(this.getDataFolder().getAbsolutePath() + getSlash() + "Existing Players.txt");
        existingPlayers = new ArrayList<>();

        if (!fTest.exists()) {
            try {
                if (!fTest.createNewFile()) {
                    sendErrorMessage(consoleSender, "Could not create Existing Players file! Aborting...");
                    return false;
                }
                PrintWriter pw = new PrintWriter(new FileWriter(fTest));
                OfflinePlayer[] players = this.getServer().getOfflinePlayers();

                for (OfflinePlayer player : players) {
                    existingPlayers.add(player.getName());
                }
                for (String player : existingPlayers) {
                    pw.println(player);
                }
                pw.close();
            } catch (Exception e) {
                e.printStackTrace();
                return false;
            }
        } else {
            try {
                BufferedReader br = new BufferedReader(new FileReader(fTest));
                String nullCheck = br.readLine();

                while (nullCheck != null) {
                    existingPlayers.add(nullCheck);
                    nullCheck = br.readLine();
                }
                br.close();
            } catch (Exception e) {
                e.printStackTrace();
                return false;
            }
        }

        return true;
    }

    public void addExistingPlayer(String playerName) {
        existingPlayers.add(playerName);

        try {
            PrintWriter pw = new PrintWriter(new FileWriter(this.getDataFolder().getAbsolutePath() + getSlash() + "Existing Players.txt", true));
            pw.println(playerName);
            pw.close();
        } catch (IOException e) {
            sendErrorMessage(consoleSender, "Failed to add " + playerName + " to Existing Players list");
        }
    }

    public ItemStack loadBook(CommandSender sender, String fileName, String isSigned, String dir) {
        return loadBook(sender, fileName, isSigned, dir, 1);
    }

    public ItemStack loadBook(CommandSender sender, String fileName, String isSigned, String dir, int numCopies) {
        //checks if player is using list number to load
        for (int x = 0; x < fileName.length(); x++) {
            if (!Character.isDigit(fileName.charAt(x))) {
                break;
            }
            if (x == fileName.length() - 1 && Character.isDigit(fileName.charAt(x)) && loadList.containsKey(sender.getName())) {
                if (Integer.parseInt(fileName) <= loadList.get(sender.getName()).size()) {
                    fileName = loadList.get(sender.getName()).get(Integer.parseInt(fileName) - 1);
                }
            }
        }

        File file;
        if (dir.equalsIgnoreCase("public")) {
            file = FileHelper.getBookFile(getBookFolder() + fileName);
        } else if (dir.equalsIgnoreCase("player")) {
            file = FileHelper.getBookFile(getBookFolder() + cleanString(sender.getName()) + getSlash() + fileName);
        } else if (dir.trim().isEmpty()) {
            file = FileHelper.getBookFile(getBookFolder() + "Encrypted" + getSlash() + dir + getSlash() + fileName);
        } else {
            file = null;
        }
        if (file == null || !file.isFile()) {
            sendErrorMessage(sender, "Incorrect file name!");
            return null;
        }

        ItemStack book;
        BookMeta bookMetadata = (BookMeta) itemFactory.getItemMeta(Material.WRITTEN_BOOK);
        if (isSigned.equalsIgnoreCase("true")) {
            book = new ItemStack(Material.WRITTEN_BOOK);
        } else {
            book = new ItemStack(Material.WRITABLE_BOOK);
        }

        BookToFromTextHelper.bookFromYml(file, bookMetadata);
        if (bookMetadata == null) {
            sendErrorMessage(sender, "File was blank!!");
            return null;
        }

        if (!dir.equalsIgnoreCase("public") && !dir.equalsIgnoreCase("player") &&
                !dir.equalsIgnoreCase("") && bookMetadata.hasLore()) {
            List<String> oldLore = bookMetadata.getLore();
            if (oldLore != null) {
                List<String> newLore = new ArrayList<>(oldLore);
                newLore.remove(0);
                bookMetadata.setLore(newLore);
            }
        }

        book.setItemMeta(bookMetadata);
        book.setAmount(numCopies);

        if (booksHavePrice() && !sender.hasPermission("bookswithoutborders.bypassBookPrice") &&
                (dir.equalsIgnoreCase("public") || dir.equalsIgnoreCase("player") ||
                        dir.equalsIgnoreCase("")) &&
                cannotPayForBookPrinting((Player) sender, numCopies)) {
            return null;
        }

        return book;
    }

    /**
     * Makes the player pay for printing a given number of books
     *
     * @param player    <p>The player printing the books</p>
     * @param numCopies <p>The number of copies the player is trying to print</p>
     * @return <p>True if the player cannot pay for the printing of the books</p>
     */
    public boolean cannotPayForBookPrinting(Player player, int numCopies) {
        //BookPriceQuantity: How many items are required to pay for each book
        //BookPriceType: Which item is used to pay for the books. AIR = use economy
        double cost = BooksWithoutBorders.bookPriceQuantity * numCopies;
        int itemCost = (int) cost;

        if (BooksWithoutBorders.bookPriceType == Material.AIR) {
            return !EconomyHelper.payForBookPrintingEconomy(player, cost, numCopies);
        } else {
            if (player.getInventory().contains(BooksWithoutBorders.bookPriceType, itemCost)) {
                payForBookPrintingItem(player, itemCost);
                return false;
            } else {
                BooksWithoutBorders.sendErrorMessage(player, itemCost + " " + BooksWithoutBorders.bookPriceType.toString() +
                        "(s) are required for this command!");
                return true;
            }
        }
    }

    /**
     * Takes payment for printing a number of books by withdrawing the correct item
     *
     * @param player   <p>The player which needs to pay</p>
     * @param itemCost <p>The number of items to pay</p>
     */
    private void payForBookPrintingItem(Player player, int itemCost) {
        PlayerInventory playerInventory = player.getInventory();

        int clearedAmount = 0;
        while (clearedAmount < itemCost) {
            int firstItemIndex = playerInventory.first(BooksWithoutBorders.bookPriceType);
            ItemStack firstItem = playerInventory.getItem(firstItemIndex);

            if (Objects.requireNonNull(firstItem).getAmount() <= itemCost - clearedAmount) {
                clearedAmount += firstItem.getAmount();
                player.getInventory().clear(firstItemIndex);
            } else {
                clearedAmount = itemCost;
                firstItem.setAmount(firstItem.getAmount() - (clearedAmount));
            }
        }
    }

    /**
     * Lists available files
     *
     * @param sender     <p>The command sender looking for files</p>
     * @param listPublic <p>Whether to list public or personal files</p>
     * @param silent     <p>Whether to just return the list without printing it</p>
     * @return <p>A list of available files</p>
     */
    public List<String> listFiles(CommandSender sender, Boolean listPublic, boolean silent) {
        File file;
        if (listPublic) {
            file = new File(getBookFolder());
        } else {
            file = new File(getBookFolder() + cleanString(sender.getName()) + getSlash());
        }
        return FileHelper.listFiles(sender, file, silent);
    }

    /**
     * Sends a success message to a command sender (player or a console)
     *
     * @param sender  <p>The sender to send the message to</p>
     * @param message <p>The message to send</p>
     */
    public static void sendSuccessMessage(CommandSender sender, String message) {
        sender.sendMessage(getSuccessColor() + message);
    }

    /**
     * Sends an error message to a command sender (player or a console)
     *
     * @param sender  <p>The sender to send the message to</p>
     * @param message <p>The message to send</p>
     */
    public static void sendErrorMessage(CommandSender sender, String message) {
        sender.sendMessage(getErrorColor() + message);
    }

    /**
     * Checks whether books have a price for printing them
     *
     * @return <p>True if players need to pay for printing books</p>
     */
    public boolean booksHavePrice() {
        return (bookPriceType != null && bookPriceQuantity > 0);
    }

}
