package net.knarcraft.bookswithoutborders.command;

import net.knarcraft.bookswithoutborders.BooksWithoutBorders;
import net.knarcraft.bookswithoutborders.utility.FileHelper;
import net.knarcraft.bookswithoutborders.utility.InputCleaningHelper;
import org.bukkit.command.Command;
import org.bukkit.command.CommandExecutor;
import org.bukkit.command.CommandSender;
import org.bukkit.entity.Player;

import java.io.File;

import static net.knarcraft.bookswithoutborders.BooksWithoutBordersSettings.getBookFolder;
import static net.knarcraft.bookswithoutborders.BooksWithoutBordersSettings.getSlash;

/**
 * Command executor for the delete command
 */
public class CommandDelete implements CommandExecutor {

    private final BooksWithoutBorders booksWithoutBorders = BooksWithoutBorders.getInstance();

    @Override
    public boolean onCommand(CommandSender sender, Command command, String label, String[] args) {
        if (!(sender instanceof Player)) {
            BooksWithoutBorders.sendErrorMessage(sender, "This command can only be used by a player!");
            return false;
        }

        return deleteBook(sender, args, false);
    }

    /**
     * Deletes a book
     *
     * @param sender       <p>The sender trying to delete the book</p>
     * @param args         <p>The arguments given</p>
     * @param deletePublic <p>Whether to delete a public book</p>
     * @return <p>True if the book was deleted successfully</p>
     */
    boolean deleteBook(CommandSender sender, String[] args, boolean deletePublic) {
        //List deletable files
        if (args.length == 0) {
            BooksWithoutBorders.loadList.put(sender.getName(), booksWithoutBorders.listFiles(sender, deletePublic, false));
            return true;
        }
        //Delete the file
        if (args.length == 1) {
            if (!BooksWithoutBorders.loadList.containsKey(sender.getName())) {
                BooksWithoutBorders.sendErrorMessage(sender, "You must first use /bwb delete to create a list of delete-able files!");
                return false;
            } else {
                if (!BooksWithoutBorders.loadList.get(sender.getName()).isEmpty()) {
                    performBookDeletion(sender, args[0], deletePublic);
                    return true;
                } else {
                    BooksWithoutBorders.sendErrorMessage(sender, "No files available to delete!");
                    return false;
                }
            }
        }
        BooksWithoutBorders.sendErrorMessage(sender, "Incorrect number of arguments for this command!");
        return false;
    }

    /**
     * Deletes the given book
     *
     * @param sender   <p>The sender of the command to delete the book</p>
     * @param fileName <p>The file name of the book</p>
     * @param isPublic <p>Whether the book to delete is public or not</p>
     */
    public void performBookDeletion(CommandSender sender, String fileName, Boolean isPublic) {
        //If the file name is an index of the load list, load the book
        try {
            int loadListIndex = Integer.parseInt(fileName);
            String senderName = sender.getName();
            if (BooksWithoutBorders.loadList.containsKey(senderName) && loadListIndex <= BooksWithoutBorders.loadList.get(senderName).size()) {
                fileName = BooksWithoutBorders.loadList.get(senderName).get(loadListIndex - 1);
            }
        } catch (NumberFormatException ignored) {
        }

        //Get the file to be deleted
        File file;
        if (isPublic) {
            file = FileHelper.getBookFile(getBookFolder() + fileName);
        } else {
            file = FileHelper.getBookFile(getBookFolder() +
                    InputCleaningHelper.cleanString(sender.getName()) + getSlash() + fileName);
        }

        //Send message if no such file could be found
        if (file == null) {
            BooksWithoutBorders.sendErrorMessage(sender, "Incorrect file name!");
            return;
        }

        //Try to delete the file
        try {
            if (file.delete()) {
                BooksWithoutBorders.sendSuccessMessage(sender, "\"" + fileName + "\" deleted successfully");
            } else {
                BooksWithoutBorders.sendErrorMessage(sender, "Deletion failed without an exception!");
            }
        } catch (Exception e) {
            BooksWithoutBorders.sendErrorMessage(sender, "Deletion failed!");
        }
    }

}
