package net.knarcraft.bookswithoutborders.command;

import net.knarcraft.bookswithoutborders.BooksWithoutBorders;
import net.knarcraft.bookswithoutborders.utility.InputCleaningHelper;
import org.bukkit.command.Command;
import org.bukkit.command.CommandExecutor;
import org.bukkit.command.CommandSender;
import org.bukkit.entity.Player;
import org.bukkit.inventory.ItemStack;

/**
 * Command executor for the give command
 */
public class CommandGive implements CommandExecutor {

    private final BooksWithoutBorders booksWithoutBorders = BooksWithoutBorders.getInstance();

    @Override
    public boolean onCommand(CommandSender sender, Command command, String label, String[] args) {
        if (!(sender instanceof Player)) {
            BooksWithoutBorders.sendErrorMessage(sender, "This command can only be used by a player!");
            return false;
        }

        return giveBook(sender, args, false, "player");
    }

    boolean giveBook(CommandSender sender, String[] args, boolean givePublic, String folder) {
        if (args.length == 1 || args.length > 4) {
            BooksWithoutBorders.sendErrorMessage(sender, "Incorrect number of arguments for this command!");
            return false;
        }

        if (args.length == 0) {
            BooksWithoutBorders.loadList.put(sender.getName(), booksWithoutBorders.listFiles(sender, givePublic, false));
            return true;
        }

        //Organize and parse input
        String bookIdentifier = args[0];
        String receivingPlayerName = args[1];
        String copies = "1";
        String isSigned = "true";
        if (args.length == 4) {
            copies = args[2];
            isSigned = args[3];
        } else if (args.length == 3) {
            if (args[2].equalsIgnoreCase("true") || args[2].equalsIgnoreCase("false")) {
                isSigned = args[2];
            } else {
                copies = args[2];
            }
        }

        //Load books available to the player
        try {
            Integer.parseInt(bookIdentifier);
            BooksWithoutBorders.loadList.put(sender.getName(), booksWithoutBorders.listFiles(sender, givePublic, true));
        } catch (NumberFormatException ignored) {
        }

        Player receivingPlayer = booksWithoutBorders.getServer().getPlayer(receivingPlayerName);
        if (receivingPlayer == null) {
            BooksWithoutBorders.sendErrorMessage(sender, "Player not found!");
            return false;
        }

        if (receivingPlayer.getInventory().firstEmpty() == -1) {
            BooksWithoutBorders.sendErrorMessage(sender, "Receiving player must have space in their inventory to receive books!");
            return false;
        }

        String bookToLoad = InputCleaningHelper.cleanString(bookIdentifier);
        try {
            ItemStack newBook = booksWithoutBorders.loadBook(sender, bookToLoad, isSigned, folder, Integer.parseInt(copies));
            if (newBook != null) {
                receivingPlayer.getInventory().addItem(newBook);
                BooksWithoutBorders.sendSuccessMessage(sender, "Book sent!");
                BooksWithoutBorders.sendSuccessMessage(receivingPlayer, "Book received!");
                return true;
            } else {
                BooksWithoutBorders.sendErrorMessage(sender, "Book failed to load!");
                return false;
            }
        } catch (NumberFormatException e) {
            BooksWithoutBorders.sendErrorMessage(sender, "Invalid number of book copies specified!");
            return false;
        }
    }

}
