package net.knarcraft.bookswithoutborders.utility;

import net.knarcraft.bookswithoutborders.BooksWithoutBorders;
import net.milkbowl.vault.economy.Economy;
import org.bukkit.Server;
import org.bukkit.entity.Player;
import org.bukkit.plugin.Plugin;
import org.bukkit.plugin.RegisteredServiceProvider;

public class EconomyHelper {

    private static Economy economy;

    /**
     * Gets an economy instance for making transactions
     * @return <p>An economy instance, or null if it's not initialized</p>
     */
    public static Economy getEconomy() {
        return economy;
    }

    /**
     * Tries to set up economy
     * @return <p>True if economy is set up and enabled</p>
     */
    public static boolean setupEconomy() {
        if (economy != null) {
            return true;
        }

        Server server = BooksWithoutBorders.getInstance().getServer();
        Plugin plugin = server.getPluginManager().getPlugin("Vault");

        if (plugin != null) {
            RegisteredServiceProvider<Economy> economyProvider = server.getServicesManager().getRegistration(net.milkbowl.vault.economy.Economy.class);
            if (economyProvider != null) {
                economy = economyProvider.getProvider();
            }
        }

        return (economy != null);
    }

    /**
     * Uses economy to take payment for printing a number of books
     *
     * @param player    <p>The player which needs to pay</p>
     * @param cost      <p>The cost of the book printing</p>
     * @param numCopies <p>The number of books the player is printing</p>
     * @return <p>True if the player had the money and it has been withdrawn</p>
     */
    public static boolean payForBookPrintingEconomy(Player player, double cost, int numCopies) {
        if ((economy.getBalance(player) - cost) >= 0) {
            economy.withdrawPlayer(player, cost);
            BooksWithoutBorders.sendSuccessMessage(player, economy.format(cost) + " withdrawn to create " + numCopies + " book(s)");
            BooksWithoutBorders.sendSuccessMessage(player, "New balance: " + economy.format(economy.getBalance(player)));
            return true;
        } else {
            BooksWithoutBorders.sendErrorMessage(player, economy.format(cost) + " is required for this command!");
            return false;
        }
    }

}
